/*
Original written by:
            Ing Hui, Cytron Technologies

Modified:
            Zamzuri, Cytron Technologies	- Fix SoftwareSerial bugs
  29/06/15  Idris, Cytron Technologies    - Point to IDE SoftwareSerial
                                          - Restructure the code style to follow standard Arduino library
*/

//#include "HardwareSerial.h"
#include "Cytron_G15Shield2.h"

boolean hardwareSerial;
SoftwareSerial* G15Serial;

// This Code is using Serial library
// baudrate 300, 1200, 2400, 4800, 9600, 14400, 19200, 28800, 38400, 57600, or 115200

Cytron_G15Shield2::Cytron_G15Shield2(uint8_t rxpin, uint8_t txpin, uint8_t ctrlpin)
{
  _rxpin = rxpin;
  _txpin = txpin;
  _ctrlpin = ctrlpin;
}

Cytron_G15Shield2::Cytron_G15Shield2(uint8_t ctrlpin)
{
	_rxpin = 0;
  _txpin = 1;
  _ctrlpin = ctrlpin;
}

void Cytron_G15Shield2::begin(uint32_t baudrate)
{
  if(_rxpin == 0 &&_txpin == 1)
  {
    hardwareSerial = true;
    Serial.begin(baudrate);
    while (!Serial);
    Serial.setTimeout(SerialTimeOut);
  }
  else
  {
    hardwareSerial = false;
    pinMode(_rxpin, INPUT);
    pinMode(_txpin, OUTPUT);
    G15Serial = new SoftwareSerial(_rxpin, _txpin);
    G15Serial->begin(baudrate);
    G15Serial->setTimeout(SerialTimeOut);
  }
  pinMode(_ctrlpin, OUTPUT);
  setTxMode();
}

void Cytron_G15Shield2::end(void)
{
  if(_rxpin == 0 &&_txpin == 1)
  {
    Serial.end();
  }
  else
  {
    pinMode(_rxpin, INPUT);
    pinMode(_txpin, INPUT);
    G15Serial->end();
  }
  pinMode(_ctrlpin, INPUT);
}

void Cytron_G15Shield2::setTxMode(void)
{
	digitalWrite(_ctrlpin, TxMode); 
}

void Cytron_G15Shield2::setRxMode(void)
{
  digitalWrite(_ctrlpin, RxMode); 
}

// Send packet
// Caution: At least 2 bytes of data array need to be passed into the function 
uint16_t Cytron_G15Shield2::sendPacket(uint8_t servoID, uint8_t instruction, uint8_t* data, uint8_t parameterLength)	
{
	uint8_t readCount = 0;
	uint8_t i;
  uint8_t packetLength = 0;
  uint8_t txBuffer[16];
  uint8_t status[16];
	uint8_t checksum = 0; // Checksum = ~(ID + Length + Instruction + Parameter1 + ... + Parameter n)
	uint16_t error = 0;
	
	setTxMode();
	
  txBuffer[0] = 0xFF; // Header is 0xFF
  txBuffer[1] = 0xFF; // Header is not included in checksum
  txBuffer[2] = servoID;              checksum += txBuffer[2]; // 0-254, 0xFE = broadcast id
  txBuffer[3] = parameterLength + 2;	checksum += txBuffer[3]; // Instruction + parameters (start add + values) + checksum                                                                                                        //0xFF and ID not included
  txBuffer[4] = instruction;          checksum += txBuffer[4];	

  for(i = 0; i < parameterLength; i++)
  {
		txBuffer[i + 5] = data[i];
		checksum += txBuffer[i + 5];
  }
  txBuffer[i+5] = ~checksum;

  packetLength = txBuffer[3] + 4; // Number of bytes for the whole packet

	if(hardwareSerial == true)
	{
		for(i = 0; i < packetLength; i++) {
			Serial.write(txBuffer[i]);
		}
		Serial.flush();
	}
	else
	{
		G15Serial->listen();
		for(i = 0; i < packetLength; i++) {
			G15Serial->write(txBuffer[i]);
		}
		G15Serial->flush();
	}
	
  // G15 only response if it was not broadcast command
  if((ID != 0xFE) || (instruction == iPING))
	{
    if(instruction == iREAD_DATA)
		{
			parameterLength = data[1]; 
      packetLength = data[1] + 6;
		}
    else {
      packetLength = 6;
		}

		setRxMode();

		if(hardwareSerial == true) {
			readCount = Serial.readBytes(status, packetLength); 
    }
		else {
			readCount = G15Serial->readBytes(status, packetLength); 
    }
		
		setTxMode();
		
		error = 0;
		if(readCount != packetLength) {
			error |= 0x0100;
		}
		if((status[0] != 0xFF) || (status[1] != 0xFF)) {
			error |= 0x0200;
		}
		if(status[2] != servoID) {
			error |= 0x0400;
		}
		if(status[4] != 0) {
			error |= status[4];
		}

		// Calculate checksum
		checksum = 0;
    for(i = 2; i < packetLength; i++) {
      checksum += status[i];
    }
    if(checksum != 0xFF) {
      error |= 0x0800;
    }
		if(status[4] == 0x00 && (error & 0x0100) == 0x00)	// Copy data only if there is no packet error
		{
			if(instruction == iPING) {
				data[0] = status[2];
			} 
			else if(instruction == iREAD_DATA)
			{
				for(i = 0; i < parameterLength; i++) {
					data[i] = status[i + 5];
        }
			}
		}
  }
  return(error);
}



//******************************************************************
//*	SET Line Mode
//* 	eg:	SetLineMode(1,iWRITE_DATA);	// Set either dark line or light line for Line Sensor of ID 1
//******************************************************************
uint16_t Cytron_G15Shield2::setLineMode(uint8_t servoID, uint8_t onOff)
{
	uint8_t txBuffer[2];

  txBuffer[0] = Dark_Light_Mode;
	txBuffer[1] = onOff; // ON = lightline, OFF = darkline

  return(sendPacket(servoID, iWRITE_DATA, txBuffer, 2));
}


//******************************************************************
//*	SET ID
//* 	eg:	SetID(MAIN,0xFE,3);	// Change the ID of any number to 3
//******************************************************************
uint16_t Cytron_G15Shield2::setID(uint8_t servoID, uint8_t newID)
{
  uint8_t txBuffer[2];
	uint16_t error; 
	
  txBuffer[0] = ID;
	txBuffer[1] = newID;	
	
	error = sendPacket(servoID, iWRITE_DATA, txBuffer, 2);
	delay(10); // Delay for eeprom write
  return(error);
}	


//******************************************************************
//*	SET S1 Threshold
//* 	eg:	SetS1Threshold(1,3);	// Change the ID of any number to 3
//******************************************************************
uint16_t Cytron_G15Shield2::setS1Threshold(uint8_t servoID, uint8_t S1threshold)
{
  uint8_t txBuffer[2];
	uint16_t error; 
	
  txBuffer[0] = S1_THRESHOLD;
	txBuffer[1] = S1threshold;	
	
	error = sendPacket(servoID, iWRITE_DATA, txBuffer, 2);
	delay(10); // Delay for eeprom write
  return(error);
}	


//******************************************************************
//*	SET S2 Threshold
//* 	eg:	SetS2Threshold(1,3);	// Change the ID of any number to 3
//******************************************************************
uint16_t Cytron_G15Shield2::setS2Threshold(uint8_t servoID, uint8_t S2threshold)
{
  uint8_t txBuffer[2];
	uint16_t error; 
	
  txBuffer[0] = S2_THRESHOLD;
	txBuffer[1] = S2threshold;	
	
	error = sendPacket(servoID, iWRITE_DATA, txBuffer, 2);
	delay(10); // Delay for eeprom write
  return(error);
}	


//******************************************************************
//*	SET S3 Threshold
//* 	eg:	SetS3Threshold(1,3);	// Change the ID of any number to 3
//******************************************************************
uint16_t Cytron_G15Shield2::setS3Threshold(uint8_t servoID, uint8_t S3threshold)
{
  uint8_t txBuffer[2];
	uint16_t error; 
	
  txBuffer[0] = S3_THRESHOLD;
	txBuffer[1] = S3threshold;	
	
	error = sendPacket(servoID, iWRITE_DATA, txBuffer, 2);
	delay(10); // Delay for eeprom write
  return(error);
}	


//******************************************************************
//*	SET S4 Threshold
//* 	eg:	SetS4Threshold(1,3);	// Change the ID of any number to 3
//******************************************************************
uint16_t Cytron_G15Shield2::setS4Threshold(uint8_t servoID, uint8_t S4threshold)
{
  uint8_t txBuffer[2];
	uint16_t error; 
	
  txBuffer[0] = S4_THRESHOLD;
	txBuffer[1] = S4threshold;	
	
	error = sendPacket(servoID, iWRITE_DATA, txBuffer, 2);
	delay(10); // Delay for eeprom write
  return(error);
}	
//******************************************************************
//*	SET Auto Calibrate
//* 	eg:	SetAutoCalibrate(1,iWRITE_DATA);	// Turn on LED of servo 2
//******************************************************************
uint16_t Cytron_G15Shield2::setAutoCalibrate(uint8_t servoID, uint8_t onOff)
{
	uint8_t txBuffer[2];

  txBuffer[0] = AUTO_CALIBRATE; // Control Starting Address
	txBuffer[1] = onOff; // ON = 1, OFF = 0

  return(sendPacket(servoID, iWRITE_DATA, txBuffer, 2));
}


//******************************************************************
//*	SET LED
//* 	eg:	SetLED(1,iWRITE_DATA);	// Turn on LED of servo 2
//******************************************************************
uint16_t Cytron_G15Shield2::setLED(uint8_t servoID, uint8_t onOff)
{
	uint8_t txBuffer[2];

  txBuffer[0] = LED; // Control Starting Address
	txBuffer[1] = onOff; // ON = 1, OFF = 0

  return(sendPacket(servoID, iWRITE_DATA, txBuffer, 2));
}


//******************************************************************
//*	SET BAUDRATE
//* 	eg:	SetBaud(1,2,1);	// Turn on torque of servo 2
//******************************************************************
uint16_t Cytron_G15Shield2::setBaudRate(uint8_t servoID, uint32_t baudrate)
{
	uint8_t txBuffer[2];
	uint16_t error; 
	
  txBuffer[0] = BAUD_RATE;
	txBuffer[1] = (2000000 / baudrate) - 1;	// Baudrate = 32M / (16*(n + 1)) = 2000000 / (n+1)
									
	error = sendPacket(servoID, iWRITE_DATA, txBuffer, 2); 
	delay(10); // Delay for eeprom write
  return(error);
}


//******************************************************************
//*	GET S1 analog value 
//******************************************************************
uint16_t Cytron_G15Shield2::getS1Analog(uint8_t servoID, uint8_t *data)
{
  data[0] = S1_AN; // Starting address where data to be read
  data[1] = 0x02; // No of bytes to be read

  return(sendPacket(servoID, iREAD_DATA, data, 2));
}


//******************************************************************
//*	GET S2 analog value 
//******************************************************************
uint16_t Cytron_G15Shield2::getS2Analog(uint8_t servoID, uint8_t *data)
{
  data[0] = S2_AN; // Starting address where data to be read
  data[1] = 0x02; // No of bytes to be read

  return(sendPacket(servoID, iREAD_DATA, data, 2));
}


//******************************************************************
//*	GET S3 analog value 
//******************************************************************
uint16_t Cytron_G15Shield2::getS3Analog(uint8_t servoID, uint8_t *data)
{
  data[0] = S3_AN; // Starting address where data to be read
  data[1] = 0x02; // No of bytes to be read

  return(sendPacket(servoID, iREAD_DATA, data, 2));
}


//******************************************************************
//*	GET S4 analog value 
//******************************************************************
uint16_t Cytron_G15Shield2::getS4Analog(uint8_t servoID, uint8_t *data)
{
  data[0] = S4_AN; // Starting address where data to be read
  data[1] = 0x02; // No of bytes to be read

  return(sendPacket(servoID, iREAD_DATA, data, 2));
}


//******************************************************************
//*	GET digital value (determine which sensor detected line)
//******************************************************************
uint16_t Cytron_G15Shield2::getDigitalVal(uint8_t servoID, uint8_t *data)
{
  data[0] = DIGITAL_VAL; // Starting address where data to be read
  data[1] = 0x01; // No of bytes to be read

  return(sendPacket(servoID, iREAD_DATA, data, 2));
}


//******************************************************************
//*	SET ACTION
//* 	eg:	SetAction(1);	// All servo action!
//******************************************************************
/*void Cytron_G15Shield2::setAction(void)
{	
	 // byte TxBuff[1];	//dummy byte
	 // sendPacket(0xFE, iACTION, TxBuff, 0);	
	set_act(_ctrlpin); 
}

void set_act(uint8_t ctrl){

	int i; 
	byte packet_len = 0;
    byte TxBuff[16];
    byte chksum = 0;	//Check Sum = ~ (ID + Length + Instruction + Parameter1 + ... Parameter N)
 
	digitalWrite(ctrl,TxMode); 

    TxBuff[0] = 0xFF;			//0xFF not included in checksum
    TxBuff[1] = 0xFF;
    TxBuff[2] = 0xFE; 			chksum += TxBuff[2];	//0-254, 0xFE = broadcast id
    TxBuff[3] =  2;	chksum += TxBuff[3];	//INSTRUCTION + PARAMETERS( START ADDR + VALUES ) + CHECKSUM
                                                                                                            //0xFF and ID not included
    TxBuff[4] = iACTION;			chksum += TxBuff[4];	//

    TxBuff[5] = ~chksum; 				//Checksum with Bit Inversion

    packet_len = TxBuff[3] + 4;			//# of bytes for the whole packet
    
	if(hardwareSerial==true)
	{
		for(i=0; i<packet_len;i++){
			Serial.write(TxBuff[i]);
		}
		Serial.flush();
	}
	else
	{
		for(i=0; i<packet_len;i++){
			G15Serial->write(TxBuff[i]);
		}
		G15Serial->flush();	
	}
}*/
