/*
Original written by:
            Ing Hui, Cytron Technologies

Modified:
            Zamzuri, Cytron Technologies	- Fix SoftwareSerial bugs
  29/06/15  Idris, Cytron Technologies    - Point to IDE SoftwareSerial
                                          - Restructure the code style to follow standard Arduino library
*/

#ifndef Cytron_G15Shield2_h
#define Cytron_G15Shield2_h

#include "Arduino.h"
#include <SoftwareSerial.h>

// Arduino Leonardo
#if defined (__AVR_ATmega32U4__)
	#define Serial Serial1
#else
	#define Serial Serial
#endif

// definitions
//******************************************************************
//*	INSTRUCTIONS
//******************************************************************
#define iPING 			0x01 // Obtain a status packet
#define iREAD_DATA	0x02 // Read Control Table values
#define iWRITE_DATA	0x03 // Write Control Table values
#define iRESET 			0x06 // Set factory defaults

#define SerialTimeOut 100L 
#define TxMode HIGH  	// Master transmit to G15
#define RxMode LOW		// Master receive from G15

enum {
	MODEL_NUMBER_L,				// 0x00
	MODEL_NUMBER_H, 				// 0x01
	F_REV, 					// 0x02
	ID, 						// 0x03
	BAUD_RATE, 				// 0x04
	RETURN_DELAY_TIME, 			// 0x05
	RETURN_ENABLE, 				// 0x06
	DARK_LIGHT_MODE, 			// 0x07
	S1_THRESHOLD, 				// 0x08
	S2_THRESHOLD, 				// 0x09
	S3_THRESHOLD, 				// 0x0A
	S4_THRESHOLD, 				// 0x0B
	ALARM, 					// 0x0C
	LED, 						// 0x0D
	S1_AN, 					// 0x0E
	S2_AN, 					// 0x0F
	S3_AN, 					// 0x10	
	S4_AN, 					// 0x11
	DIGITAL_VAL, 				// 0x12
	RESERVED, 					// 0x13
	AUTO_CALIBRATE, 				// 0x14
	LOCK, 					// 0x15
	};

//Hvoid set_act(char ctrl);  

class Cytron_G15Shield2
{
	public:
		uint8_t ServoID;
		uint8_t _txpin, _rxpin, _ctrlpin;
		
		Cytron_G15Shield2(uint8_t rxpin, uint8_t txpin, uint8_t ctrlpin); // SoftwareSerial
		Cytron_G15Shield2(uint8_t ctrlpin); // HardwareSerial
		void begin(uint32_t baudrate);
		void end(void);
		
		//*=========Wheel Mode=====================================================================================
		//360 degree continous rotation. change CW and CCW Angle Limits to same value
		uint16_t setWheelMode(uint8_t servoID);
		uint16_t exitWheelMode(uint8_t servoID);
		uint16_t setWheelSpeed(uint8_t servoID, uint16_t speed, uint8_t direction);

		//*=========Normal Positioning Mode========================================================================
		//(Rotation limited by Angle Limit and Direction of Rotation determined by operation section of Angle Limit)
		uint16_t setPos(uint8_t servoID, uint16_t position);
		uint16_t setPosAngle(uint8_t servoID, uint16_t angle);
		uint16_t setPosSpeed(uint8_t servoID, uint16_t position, uint16_t speed);
		
		//*========Direction Positioning Mode======================================================================
		//(Rotation direction and angle is NOT limited by Angle Limit Control Register value)
		uint16_t rotateCW(uint8_t servoID, uint16_t position);
		uint16_t rotateCCW(uint8_t servoID, uint16_t position);
		
		//*=======Torque Enable and Speed Control==================================================================
		uint16_t setTorqueOnOff(uint8_t servoID, uint8_t onOff);
		uint16_t setSpeed(uint8_t servoID, uint16_t speed);
		uint16_t setTimeToGoal(uint8_t servoID, uint16_t time);
		
		//*=======Set Maximum Limits===============================================================================
		uint16_t setAngleLimit(uint8_t servoID, uint16_t cwAngle, uint16_t ccwAngle);
		uint16_t setTorqueLimit(uint8_t servoID, uint16_t torqueLimit); //in RAM area
		uint16_t setTemperatureLimit(uint8_t servoID, uint8_t temperature);
		uint16_t setVoltageLimit(uint8_t servoID, uint8_t voltageLow, uint8_t voltageHigh);
		
		uint16_t setID(uint8_t servoID, uint8_t newID);
		
		uint16_t setLED(uint8_t servoID, uint8_t onOff);
		uint16_t setAlarmLED(uint8_t servoID, uint8_t alarmLED);
		uint16_t setAlarmShutDown(uint8_t servoID, uint8_t alarm);
		
		//*========Servo Positioning Control Parameters============================================================
		uint16_t setMarginSlopePunch(uint8_t servoID, uint8_t CWMargin, uint8_t CCWMargin, uint8_t CWSlope, uint8_t CCWSlope, uint16_t punch);

		uint16_t setBaudRate(uint8_t servoID, uint32_t baudrate);
		uint16_t factoryReset(uint8_t servoID);
		uint16_t ping(uint8_t servoID, uint8_t *data);
		
		uint16_t getPos(uint8_t servoID, uint8_t *data);
		uint16_t getSpeed(uint8_t servoID, uint8_t *data);
		uint16_t getLoad(uint8_t servoID, uint8_t *data); 
		uint16_t getVoltage(uint8_t servoID, uint8_t *data); 
		uint16_t getTemperature(uint8_t servoID, uint8_t *data);
		uint16_t getTorqueOnOff(uint8_t servoID, uint8_t *data);		
		uint16_t isMoving(uint8_t servoID, uint8_t *data);
		
		//static void setAction(void);

	protected:
		void setRxMode(void);
		void setTxMode(void);
		uint16_t sendPacket(uint8_t servoID, uint8_t instruction, uint8_t* data, uint8_t parameterLength);
		//byte read_data(byte id, byte* data);
};

#endif
